'Filename   SWRMeter_Mega8_110.bas
'Version    1.10
'Date       2013-09-11
'Purpose    SWR and Power meter with
'           - mathematical approximation of the diode U/I-characteristic
'             to calculate the 'real' HF voltage from measured diode DC voltage
'             valid for Schotty BAT43.
'             For other diodes pls. measure the characteristic and
'             fit the data to the function y = a * x^b + 1, x = DC voltage
'             with the help of the Excel function RKP.
'           - display of SWR & RF foreward/reverse power
'           - Fine tuning can be done with gain adjustment on OpAmps.
'             For reliable power calibration pls. use a JFET/CMOS opamp
'             with adjustable 1.5 ... 2x gain, e.g. TLC272.
'           - Choose toroid turns of tandem match coupler fitting max. RF power
'             Change (see below): bytTurns = 9, fits 20W
'           - SWR bridge data are set in SWR_bridge_data.bas
'             wrdVoltRef = 4982     ADC supply/reference voltage AVCC (mV)
'             bytTurns = 9          Power calibration factor
'                                   =turns of tandem match coupler toroids
'             sngDiodeA = 0.251607  Diode fitparameter "a"
'             sngDiodeB = -0.857977 Diode fitparameter "b"
'             sngDiodeC = 1         Diode fitparameter "c"
'             wrdDiodeLow = 20      lower Diode detection limit (mV)
'Author     Georg Latzel, DL6GL
'Copyright  (c) Georg Latzel, free for private use under GNU license
'Hardware   ATmega8
'           Crystal 16 MHz
'           Port B0: Jumper or switch eff./peak HF power
'           Port B1: LED1 eff./peak HF power, lit when Port B0 is grounded
'           Port B2: LED2 is lit when SWR > 3
'           Ports PC0, PC1 (ADC0, ADC1): SWR Fwd, Rev
'           Ports D.0 ... D5: 2x16 Text LCD

'Compiler   BASCOM-AVR 2.0.7.3
'           For readable "camel case" variables mark "Don't change case"
'           in BASCOM Options-Programmer-Environment-Editor
'           Flash used: 5,530 bytes (67%)
'Edit history
'V1.00      2012-08-04, first release
'V1.10      2013-09-11
'           BASCOM "GETADC" replaced by ADC register handling (faster)
'           Added: Jumper/switch for HF eff/peak power display
'           Added: LED1 to show HF eff/peak power setting
'           Added: LED2: SWR > 3 alert
'           SWR bridge data set and included via SWR_Bridge_Data.bas

'==============================================================================

'FUSE-Bits (AVR Studio 4.0-Settings, [ ]=no check, [x]=check)
'PonyProg does'nt know the ATmega1284P
'In ATMEGA-Data sheets: Bit = 1 -> not set
'                       Bit = 0 -> set
'BODLEVEL   disabled  (BODLEVEL=111)
'OCDEN      [ ]
'JTAGEN     [ ] (disabled)
'SPIEN      [x] Do not change!
'WDTON      [ ]
'EESAVE     [ ]
'BOOTSZ     Boot Flash size=512 words (minimal boot size)
'BOOTRST    [ ]
'CKDIV8     [ ]
'CKOUT      [ ]
'SUT_CKSEL  Ext. Crystal Osc. 8.0- MHz, Start-up time 16k + 65ms

$regfile = "m8def.dat"                            'ATMega8
$crystal = 16000000                               '16 MHz xtal
$hwstack = 64
$swstack = 64
$framesize = 64

'Customizing variables for ADC measurement ------------------------------------
'System impedance sngImp, usually sngImp = 50 Ohm
'Max. SWR alert sngSWRMax, LED2 is lit
Const sngImp = 50                                 'System impedance
Const sngSWRMax = 3                               'Max. SWR: LED2 is lit

'Diode response parameters ----------------------------------------------------
'Best fit for Schottky diode correction function
'CF = a * DC^b + c
'DC = measured rectified DC voltage
'HF = CF * DC, best fit for the "real" HF voltage
'Best fit parameters for Schottky BAT43:
Dim wrdDiodeLow As Word                           'lower Diode detection limit (mV)
Dim sngDiodeLow As Single                         'lower Diode detection limit (V)
Dim sngDiodeA As Single                           'Parameter "a"
Dim sngDiodeB As Single                           'Parameter "b"
Dim sngDiodeC As Single                           'Parameter "c"

'Configure in/out on Port B ---------------------------------------------------
'J1   Jumper RF power      PB0
'LED1 RF power eff/peak    PB1
'LED2 Power > 5W           PB2
DdrB.0 = 0                                        'PB.0=Input Jumper J1
PortB.0 = 1                                       'Activate pullup @ PortB.0
DDRB.1 = 1                                        'PB.1=Output LED1
DDRB.2 = 1                                        'PB.2=Output LED2
Jumper1 Alias PinB.0                              'Input Jumper RF eff/peak
LED1 Alias PortB.1                                'Output LED1 @ PortB.1
LED2 Alias PortB.2                                'Output LED2 @ PortB.2

'ADC on Port C ----------------------------------------------------------------
'ADC0                      PC0                    'Forward voltage
'ADC1                      PC1                    'Reverse voltage
'ADC2                      PC2                    'Not used
'ADC3                      PC3                    'Not used
'ADC4                      PC4                    'Not used
'ADC5                      PC5                    'Not used
DdrC = &B00000000                                 'PortC = input
PortC = &B11111100                                'Pullup PC2...PC7

'Configure LCD on Port D ------------------------------------------------------
'LCD     RS                PD0
'LCD     E                 PD1
'LCD     D4                PD2
'LCD     D5                PD3
'LCD     D6                PD4
'LCD     D7                PD5

Config Lcd = 16 * 2
Config Lcdpin = Pin , Db7 = PortD.5 , Db6 = PortD.4 , Db5 = PortD.3
Config Lcdpin = Pin , Db4 = PortD.2 , E = PortD.1 , Rs = PortD.0
Config Lcdbus = 4

Const bytRowPwr = 1                               'Row for RF power display
Const bytRowSWR = 2                               'Row for SWR display
Const bytChar = 16                                'Characters per LCD line
Const bytDots = 5                                 'bar dots per character
Const bytBarLen = 12                              'Bar length (characters)

'Define display bars ----------------------------------------------------------
Deflcdchar 0 , 16 , 16 , 16 , 16 , 16 , 16 , 16 , 32       'chr(0): 1 bar
Deflcdchar 1 , 24 , 24 , 24 , 24 , 24 , 24 , 24 , 32       'chr(1): 2 bars
Deflcdchar 2 , 28 , 28 , 28 , 28 , 28 , 28 , 28 , 32       'chr(2): 3 bars
Deflcdchar 3 , 30 , 30 , 30 , 30 , 30 , 30 , 30 , 32       'chr(3): 4 bars
Deflcdchar 4 , 31 , 31 , 31 , 31 , 31 , 31 , 31 , 32       'chr(4): 5 bars (full)

'Common Variables, multiply used ----------------------------------------------
Dim bytTmp0 As Byte
Dim wrdTmp0 As Word
Dim wrdTmp1 As Word
Dim wrdTmp2 As Word
Dim sngTmp0 As Single
Dim sngTmp1 As Single
Dim strTmp6 As String * 6
Dim strDis As String * 16                         'Text to be dispayed on LCD
Dim bytDis(16) As Byte At strDis Overlay          'the same as byte array

'ADC Variables ----------------------------------------------------------------
Dim bytADMUX As Byte
Dim wrdADC As Word
Dim bytADCLo As Byte At wrdADC Overlay
Dim bytADCHi As Byte At wrdADC + 1 Overlay
Dim bitADCdone As Bit                             '1= all ADC channels complete
Dim bitADCPWR As Bit                              '=1: Power accumulation done
Dim bytADCRun As Byte                             'Current ADC run for mean value
Dim wrdADCVal(2) As Word                          '1=Forward, 2=Reverse
Dim bytADChannel As Byte                          '0=Forward, 1=Reverse
Dim bytADCValIdx As Byte                          'Current ADC value index
Dim bytPWRerr As Byte
Const bytADCRep = 3                               'Repetitive ADC runs for mean value
Const bytADCUsed = 1                              'Number of used ADC channels - 1

'SWR & power Variables --------------------------------------------------------
Dim bytTurns As Byte                              'Power calibration factor
Dim wrdVoltRef As Word                            'DC supply/reference voltage AVCC (mV)
Dim sngADCFwd As Single                           'ADC forward voltage
Dim sngADCRev As Single                           'ADC reverse voltage
Dim sngFwd As Single                              'Mean ADC forward voltage
Dim sngRev As Single                              'Mean ADC reverse voltage
Dim sngSWR As Single                              'SWR
Dim sngADCCal As Single                           '10 bit ADC calibration
Dim sngSWRBarCal As Single                        'SWR bar scale factor
Const sngSWRBarMax = 5                            'Max SWR for bargraph

'Declare subs & functions -----------------------------------------------------
Declare Sub ClearLCDLine(byVal bytRow As Byte , byVal bytChars As Byte)
Declare Sub GetADCData
Declare Sub ADCalibrate
Declare Function DiodeResponse(byVal sngDC As Single) As Single
Declare Sub ShowPower
Declare Sub ShowSWR

'Configure ADC, Reference = AVCC (ATmega8/16/32) ------------------------------
bytADMUX = &B01000000                             'Reference = AVCC, ADLAR=0, ADC0
Admux = bytADMUX                                  'ADC Multiplexer Selection Register
'                                                  Bit 7 , 6 : REFS1 , REFS0
'                                                  00: Reference = AREF
'                                                  01: Reference = AVCC
'                                                  10: Int. 1.10V Reference
'                                                  11: Int. 2.56V Reference
'                                                  Bit 5, ADLAR=0: data right adjusted
'                                                              =1: data left adjusted
'                                                  Bit 4-0, MUX4...MUX0
'                                                  =0000: ADC0 is input
'                                                  =0001: ADC1 is input
'                                                  =0010: ADC2 is input
'                                                  =0011: ADC3 is input
'                                                  =0100: ADC4 is input
'                                                  =0101: ADC5 is input
'                                                  =0110: ADC6 is input
'                                                  =0111: ADC7 is input
ADCSRA = &B10001111                               'Enable ADC, div. factor 128
'                                                  ADC Control & Status Register A
'                                                  Bit 7, ADEN = 1: enable ADC
'                                                  Bit 6, ADSC = 1: Start ADC
'                                                  Bit 5, ADATE = 1: Auto trigger enabled
'                                                  Bit 4, ADIF = 1 when conversion completed
'                                                  Bit 3, ADIE = 1: Conversion completion interrupt activated
'                                                  Bit 2-0, ADPS2...ADPS0, ADC prescaler
'                                                  =000: Div. factor 2
'                                                  =001: Div. factor 2
'                                                  =010: Div. factor 4
'                                                  =011: Div. factor 8
'                                                  =100: Div. factor 16
'                                                  =101: Div. factor 32
'                                                  =110: Div. factor 64
'                                                  =111: Div. factor 128
'SFIOR = &B00000000                                '&B00000000 is the default

On ADC ADC_isr                                    'ADC conversion complete interrupt
SREG.7 = 1                                        'Enable all interrupts

Cls
Cursor Off
Call ClearLCDLine(1 , bytChar)
LCD "SWR-Meter 1.10"
Call ClearLCDLine(2 , bytChar)
LCD "DL6GL 2013-09-11"
Wait 2
Cls

'Initialize data --------------------------------------------------------------
$Include SWR_Bridge_Data.bas                      'Initialize Bridge data
sngADCCal = wrdVoltRef                            'ADC Reference voltage (mV)
sngADCCal = 0.001 * sngADCCal                     'ADC Reference voltage (V)
sngADCCal = sngADCCal / 1024                      'ADC voltage calibration
sngDiodeLow = wrdDiodeLow                         'lower Diode detection limit (mV)
sngDiodeLow = 0.001 * sngDiodeLow                 'lower Diode detection limit (V)

sngADCFwd = 0                                     'ADC forward voltage
sngADCRev = 0                                     'ADC reverse voltage
bytADCRun = 0
bytADChannel = 0                                  'First ADC channel ADC0
bitADCdone = 0
sngSWRBarCal = bytBarLen * bytDots                'SWR bar dots (5 per character)
sngSWRBarCal = sngSWRBarCal / sngSWRBarMax        'SWR bar calibration
ADCSRA.ADSC = 1                                   'Start ADC
Waitms 500

Do                                                'Start of main program ******

   If Jumper1 = 0 Then                            'Jumper1 in: eff. power measured
      LED1 = 1                                    'Turn LED1 on, eff. power
   Else
      LED1 = 0                                    'Turn LED1 off, peak power
   End If

   Call GetADCData                                'Accumulate ADC values
   If bitADCPWR = 1 Then                          'ADC data accumulated
      bytADCRun = 0                               'Initialize next ADC run
      Call ADCalibrate                            'Calibrate ADC values
      Select Case bytPWRerr
         Case 0                                   'OK, calculate SWR
            Call ShowPower                        'Show power
            Call ShowSWR                          'Show SWR in row bytRowSWR
         Case 1
               Locate bytRowSWR , 1
               LCD "Err: Signal low "
         Case 2
            Locate bytRowSWR , 1
            LCD "Err: FWD < REV  "
         Case 3
            Locate bytRowPwr , 1
            LCD "No RF signal    "
      End Select
      sngADCFwd = 0                               'Initialize for next ADC run
      sngADCRev = 0
   End If

Loop                                              'End of main program *******

End

'==============================================================================
ADC_isr:
   'Interrupt Service Routine for ADC, conversion complete
   'Read 10bit ADC Data Registers ADCL & ADCH
   'bytADCLo and bytADCHi are overlayed with wrdADC
   'Set next MUX channel and start ADC again

   bytADCLo = ADCL                                'Read ADCL first...
   bytADCHi = ADCH                                '... then ADCH
   bytADCValIdx = bytADChannel + 1                'Next value array position
   wrdADCVal(bytADCValIdx) = wrdADC               '10bit ADC value
   Incr bytADChannel                              'Next MUX channel
   If bytADChannel > bytADCUsed Then
      bytADChannel = 0
      bitADCdone = 1                              'Accumulation complete
   End If
   Admux = bytADMUX Or bytADChannel               'Set next MUX channel
   ADCSRA.ADSC = 1                                'Start ADC again
Return

'==============================================================================
Sub GetADCData
'Get accumulated ADC Data
'Input:  wrdADCVal(1)   ADC forward voltage
'        wrdADCVal(2)   ADC reverse voltage
'        bytADCRep      Number of repetitive ADC accumulations
'Output: sngADCFwd      Mean ADC forward voltage
'        sngADCRev      Mean ADC reverse voltage
'        bytADCRun      Current ADC accumulation number

   If bitADCdone = 1 Then                         'ADC accumulation complete
      If bytADCRun < bytADCRep Then               'Accumulate bytADCRep times
         sngADCFwd = sngADCFwd + wrdADCVal(1)     'ADC RF peak power value
         sngADCRev = sngADCRev + wrdADCVal(2)     'ADC Supply voltage
         Incr bytADCRun
         bitADCPWR = 0                            'ADC accumulating
      Else                                        'Mean values complete
         sngFwd = sngADCFwd / bytADCRep           'Mean values
         sngRev = sngADCRev / bytADCRep
         bitADCPWR = 1                            'ADC accumulation done
      End If
      bitADCdone = 0
   End If

End Sub

'==============================================================================
Sub ADCalibrate
'Calibrate ADC for power/SWR measurement
'Input:  sngFWD      Forward ADC count
'        sngREV      Reverse ADC count
'        sngDiodeLow Diode threshold value
'Output  sngFWD      Forward calibrated voltage
'        sngREV      Reverse calibrated voltage
'        bitTXOn     = ON: TX is on, = OFF: TX is off
'        bytPWRerr   Error flag
'                    = 0: no error
'                    = 1: signal low
'                    = 2: FWD < REV
'                    = 3: no signal
'Function DiodeResponse for correction of diode response

   If sngFWD > 5 Then                             'TX is on (Fwd >50 mV)
      If sngFWD > sngREV Then                     'FWD must be > REV
         sngFWD = sngFWD * sngADCCal              'Forward peak voltage
         sngREV = sngREV * sngADCCal              'Reverse peak voltage
         If sngFWD > sngDiodeLow Then             'Signal level OK
            sngFWD = DiodeResponse(sngFWD)        'Diode correction forward
            sngREV = DiodeResponse(sngREV)        'Diode correction reverse
            bytPWRerr = 0                         'Calibration done
         Else
            bytPWRerr = 1                         'Signal low
         End If
      Else
         bytPWRerr = 2                            'FWD < REV
      End If
   Else
      bytPWRerr = 3                               'Fwd < 25 mV, TX must be off
   End If

End Sub

'==============================================================================
Function DiodeResponse(byVal sngDC As Single) As Single
'Calculate "real" HF voltage from measured DC voltage sngDC
'Formula valid for Schotty BAT43 diode peak detector
'Correction factor:  CF = a * DC^b + c
'"Real" HF:          HF = CF * DC voltage
'Input:  sngDC, Diode DC voltage (Forward or Reverse)
'        sngDiodeA, ...B, ...C, best fit parameters for correction function
'Output: HF voltage from best fit correction function

   sngTmp0 = sngDiodeB                            'Const is not accepted by Power
   sngTmp1 = Power(sngDC , sngTmp0)
   sngTmp1 = sngTmp1 * sngDiodeA
   sngTmp1 = sngTmp1 + sngDiodeC                  'Correction factor
   DiodeResponse = sngDC * sngTmp1                'Multiply with DC voltage

End Function

'==============================================================================
Sub ShowPower
'Calculate an show forward & reverse RF Power
'Power   = (voltage)^2 / impedance
'eff voltage  = peak voltage / SQR(2)
'Input:  sngFWD      forward peak voltage
'        sngREV      reverse peak voltage
'        sngImp      circuit impedance
'        sngTurns    Voltage calibration factor (toroid turns ratio)

   sngTmp0 = sngFWD                               'Peak forward voltage
   sngTmp1 = sngREV                               'Peak reverse voltage
   If Jumper1 = 0 Then                            'Jumper in: effective power
      sngTmp0 = 0.7071 * sngTmp0                  'eff. forward voltage
      sngTmp1 = 0.7071 * sngTmp1                  'eff. reverse voltage
   End If
   sngTmp0 = sngTmp0 * bytTurns                   'Voltage calibration
   sngTmp1 = sngTmp1 * bytTurns
   sngTmp0 = sngTmp0 * sngTmp0                    'forward voltage squared
   sngTmp1 = sngTmp1 * sngTmp1                    'reverse voltage squared
   sngTmp0 = sngTmp0 / sngImp                     'forward power
   sngTmp1 = sngTmp1 / sngImp                     'reverse power

   wrdTmp0 = sngTmp0 * 10                         'forward power, show 1 decimal
   wrdTmp1 = sngTmp1 * 10                         'reverse power, show 1 decimal

   strTmp6 = Str(wrdTmp0)
   strTmp6 = Format(strTmp6 , "  0.0")
   strDis = "F " + strTmp6                        'forward power ...
   strTmp6 = Str(wrdTmp1)
   strTmp6 = Format(strTmp6 , "  0.0")
   strDis = strDis + " R " + strTmp6 + "W"        '... and reverse power
   Locate bytRowPwr , 1
   LCD strDis                                     'Show me

 End Sub

'==============================================================================
Sub ShowSWR
'Paint value sngVal on LCD, Row bytRow, as bar graph
'Input:  sngFWD         Foreward voltage
'        sngREV         Reverse  voltage
'        sngSWRMax      max. SWR to be displayed (end of bar)
'        sngSWRBarCal   calibration factor (dots / max value)

   sngTmp0 = sngFWD + sngREV
   sngTmp1 = sngFWD - sngREV
   sngSWR = sngTmp0 / sngTmp1                     'SWR
   If sngSWR > 9.99 Then
      sngSWR = 9.99                               'SWR display limit
   End If
   If sngSWR > sngSWRMax Then                     'Max SWR exeeded?
      LED2 = 1                                    'Yes, LED2 on
   Else
      LED2 = 0                                    'No, LED2 off
   End If

   sngTmp0 = sngSWR
   If sngTmp0 > sngSWRBarMax Then
      sngTmp0 = sngSWRBarMax                      'Display limit
   End If
   sngTmp0 = sngSWRBarCal * sngTmp0               'Bar calibration
   wrdTmp2 = Int(sngTmp0)
   sngTmp0 = sngTmp0 / bytDots
   wrdTmp0 = Int(sngTmp0)                         'Number of 5 dot full blocks
   wrdTmp1 = wrdTmp0 * bytDots
   wrdTmp1 = wrdTmp2 - wrdTmp1                    'Remaining partial bars

   strDis = ""
   If wrdTmp0 > 0 Then                            'Handle full 5 dot blocks
      strDis = String(wrdTmp0 , 4)                'Fill with Chr(4)
   End If

   Select Case wrdTmp1                            'Add last bar(s) < 5 dots
      Case 1
           strDis = strDis + Chr(0)               '1 bar
      Case 2
           strDis = strDis + Chr(1)               '2 bars
      Case 3
           strDis = strDis + Chr(2)               '3 bars
      Case 4
           strDis = strDis + Chr(3)               '4 bars
   End Select

   wrdTmp1 = bytBarLen - 4                        'Space for Tag 'SWR'?
   If wrdTmp0 < wrdTmp1 Then
      bytTmp0 = 1                                 'yes
   Else
      bytTmp0 = 0                                 'no
   End If

   wrdTmp0 = Len(strDis)                          'string length up to now
   wrdTmp1 = bytChar - wrdTmp0                    'remaining characters
   wrdTmp1 = wrdTmp1 - 4                          '4 char's for value
   If wrdTmp1 > 0 Then
      strDis = strDis + Space(wrdTmp1)            'Fill with spaces
   End If
   If bytTmp0 = 1 Then                            'Add tag 'SWR'
      wrdTmp0 = Len(strDis)
      wrdTmp1 = wrdTmp0 - 4
      strDis = Left(strDis , wrdTmp1)
      strDis = strDis + "SWR "
   End If

   wrdTmp0 = sngSWR * 100                         'show 2 decimals
   strTmp6 = str(wrdTmp0)
   strTmp6 = Format(strTmp6 , "0.00")
   strDis = strDis + strTmp6                      'Add sngSWR value at the end
   Locate bytRowSWR , 1
   LCD strDis                                     'Show me

End Sub

'==============================================================================
Sub ClearLCDLine(byVal bytRow As Byte , byVal bytChars As Byte)
'Clear LCD bytChars characters in line bytRow
   Locate bytRow , 1
   LCD SPC(bytChars)
   Locate bytRow , 1
End Sub

'==============================================================================